open Elang
open Prog
open Utils
open Batteries
open BatList

type cfg_arg = 
|Var of string
|Addr of int * int (*Addr(ofs,sz)*)

type expr =
    Ebinop of binop * expr * expr
  | Eunop of unop * expr
  | Eint of int
  | Evar of string
  | Ecall of string * expr list
  | Estk of int
  | Eload of expr * int


type cfg_node =
  | Cassign of string * expr * int
  | Creturn of expr
  | Cprint of expr * int
  | Ccmp of expr * int * int (*si expr then int1 else int2*)
  | Cnop of int
  | Ccall of string * expr list * int
  | Cstore of expr * expr * int * int

type cfg_fun = {
  cfgfunargs: cfg_arg list;
  cfgfunbody: (int, cfg_node) Hashtbl.t;
  cfgentry: int;
  cfgfunstksz : int;
}

type cprog = cfg_fun prog


(* [succs cfg n] donne l'ensemble des successeurs d'un nœud [n] dans un CFG
   [cfg]. *)
let succs cfg n =
  match Hashtbl.find_option cfg n with
  | None -> Set.empty
  | Some (Cprint (_, s)) -> Set.singleton s
  | Some (Cassign (_, _, s)) -> Set.singleton s
  | Some (Creturn _) -> Set.empty
  | Some (Ccmp (_, s1, s2)) -> Set.of_list [s1;s2]
  | Some (Cnop s) -> Set.singleton s
  | Some (Ccall(_,_,s)) -> Set.singleton s
  | Some (Cstore(_,_,_,s)) -> Set.singleton s


(* [preds cfg n] donne l'ensemble des prédécesseurs d'un nœud [n] dans un CFG [cfg]
   *)
let preds cfgfunbody n =
  Hashtbl.fold (fun ind_node node acc ->
      match node with
      | Cassign (_, _, s) -> if s = n then Set.add ind_node acc else acc
      | Cprint (_, s) -> if s = n then Set.add ind_node acc else acc
      | Cnop s -> if s = n then Set.add ind_node acc else acc
      | Creturn _ -> acc
      | Ccmp (_, s1, s2) -> if s1 = n || s2 = n then Set.add ind_node acc else acc
      | Ccall(_,_ ,s) -> if s = n then Set.add ind_node acc else acc
      | Cstore(_,_,_,s) -> if s = n then Set.add ind_node acc else acc
    ) cfgfunbody Set.empty


let size_binop b e1 e2 =
  1 + e1 + e2

let size_unop u e =
  1 + e

let size_fcall f l =
  1 + (List.fold_left (fun acc elt -> acc + elt) 0 l)

  let rec size_expr (e: expr) : int =
  match e with
  | Ebinop (b, e1, e2) -> size_binop b (size_expr e1) (size_expr e2)
  | Eunop (u, e) -> size_unop u (size_expr e)
  | Eint _ -> 1
  | Evar v -> 1
  | Ecall (fname,l_expr) -> size_fcall fname (List.map (fun eXpr -> size_expr eXpr) l_expr) 
  | Estk _ -> 1
  | Eload(e,s) -> 1 + size_expr e
let rec size_instr (i: cfg_node) : int =
  match (i : cfg_node) with
  | Cassign (v, e, s) -> 1 + size_expr e
  | Creturn e -> 1 + (size_expr e)
  | Cprint (e, s) -> 1 + (size_expr e)
  | Ccmp (e, s1, s2) -> 1 + size_expr e
  | Cnop s -> 1
  | Ccall(fname,l_expr,s) -> size_expr (Ecall(fname,l_expr))
  | Cstore(e1,e2,_,_) -> 1 + (size_expr e1) + (size_expr e2)
let size_fun f =
  Hashtbl.fold (fun k v acc -> acc + size_instr v) f 0

