open Batteries
open Elang
open Elang_gen
open Cfg
open Utils
open Prog
open Report
open Cfg_print
open Options

(* [cfg_expr_of_eexpr e] converts an [Elang.expr] into a [expr res]. This should
   always succeed and be straightforward.

   In later versions of this compiler, you will add more things to [Elang.expr]
   but not to [Cfg.expr], hence the distinction.
*)
let rec cfg_expr_of_eexpr (f : Elang.efun) (typ_fun : (string, typ list * typ) Hashtbl.t) (e: Elang.expr) : expr res =
  match e with
  
  | Elang.Ebinop (binop, e1, e2) -> (* pb operations sur pointeur non traité*)
    cfg_expr_of_eexpr f typ_fun e1 >>= fun ee1 ->
    cfg_expr_of_eexpr f typ_fun e2 >>= fun ee2 ->
    type_expr f.funvartyp typ_fun e1 >>= fun t1 ->
    type_expr f.funvartyp typ_fun e2 >>= fun t2 ->
    begin match t1,t2 with 
        |Tptr _,Tint when binop = Eadd-> OK(Ebinop(Esub,ee1,(Ebinop(Emul,Eint(size_type t1),ee2))))

        |Tint,Tptr _ when binop =Eadd ->  OK(Ebinop(Esub,ee2,(Ebinop(Emul,Eint(size_type t2),ee1))))

        |Tptr _,Tint when binop = Esub-> OK(Ebinop(Eadd,ee1,(Ebinop(Emul,Eint(size_type t1),ee2))))
        |_,_  -> OK (Ebinop (binop, ee1, ee2))

    end
    
    
  
  | Elang.Eunop (u, e) ->
  cfg_expr_of_eexpr f typ_fun e >>= fun ee ->
  OK (Eunop (u, ee))

  | Elang.Eint i -> OK (Eint i)

  | Elang.Echar(c) -> OK(Eint(Char.code(c)))

  | Elang.Evar v -> (*a variable can either be in env or in mem*)
    (match Hashtbl.find_option f.funvarinmem v with
      |None -> OK(Evar(v))
      |Some(ofs) -> OK(Eload(Estk(ofs),size_type(Hashtbl.find f.funvartyp v))))
    

  | Elang.Ecall(fname,l_expr) -> list_map_res (fun eXpr -> cfg_expr_of_eexpr f typ_fun eXpr) l_expr >>= fun l -> OK(Ecall(fname,l))

  | Elang.Eaddrof(e) ->
    cfg_expr_of_eexpr f typ_fun e >>= fun e ->
    begin match e with
      | Eload(Estk(ofs),_) -> OK(Estk(ofs))  (*a variable has to be in memory*)
      |_ -> Error (Format.sprintf "Can only calculate the address of a varibale, got %s" (dump_cfgexpr e)) 
      end

  | Elang.Eload(e') ->
    cfg_expr_of_eexpr f typ_fun e' >>= fun e' ->
      type_expr f.funvartyp typ_fun e >>= fun t ->
      OK(Cfg.Eload(e',size_type t))



(* [cfg_node_of_einstr next cfg succ i] builds the CFG node(s) that correspond
   to the E instruction [i].

   [cfg] is the current state of the control-flow graph.

   [succ] is the successor of this node in the CFG, i.e. where to go after this
   instruction.

   [next] is the next available CFG node identifier.

   This function returns a pair (n, next) where [n] is the identifer of the
   node generated, and [next] is the new next available CFG node identifier.

   Hint: several nodes may be generated for a single E instruction.
*)
let rec cfg_node_of_einstr (f : Elang.efun) (typ_fun : (string, typ list * typ) Hashtbl.t) (next: int) (cfg : (int, cfg_node) Hashtbl.t)
    (succ: int) (i: instr) : (int * int) res =
  match i with
  
  | Elang.Iassign (name_var, e) ->
    cfg_expr_of_eexpr f typ_fun e >>= fun e ->
    (match Hashtbl.find_option f.funvarinmem name_var with
    |Some(ofs) -> Hashtbl.replace cfg next (Cstore(Estk(ofs),e,size_type(Hashtbl.find f.funvartyp name_var),succ)) ; OK (next, next + 1)
    |None -> 
      Hashtbl.replace cfg next (Cassign(name_var,e,succ));
      OK (next, next + 1))

  | Elang.Iif (c, ithen, ielse) ->
  cfg_expr_of_eexpr f typ_fun c >>= fun c ->
  cfg_node_of_einstr f typ_fun next cfg succ ithen >>= fun (nthen, next) ->
  cfg_node_of_einstr f typ_fun next cfg succ ielse  >>= fun (nelse, next) ->
  Hashtbl.replace cfg next (Ccmp(c, nthen, nelse)); OK (next, next + 1)

  | Elang.Iwhile (c, i) ->
  cfg_expr_of_eexpr f typ_fun c >>= fun c ->
  let (cmp, next) = (next, next+1) in
  cfg_node_of_einstr f typ_fun next cfg cmp i >>= fun (nthen, next) ->
  Hashtbl.replace cfg cmp (Ccmp(c, nthen, succ)); OK (cmp, next + 1)

  | Elang.Iblock il ->
  List.fold_right (fun i acc ->
      acc >>= fun (succ, next) ->
      cfg_node_of_einstr f typ_fun next cfg succ i
    ) il (OK (succ, next))

  | Elang.Ireturn e ->
  cfg_expr_of_eexpr f typ_fun e >>= fun e ->
  Hashtbl.replace cfg next (Creturn e); OK (next, next + 1)
  
  | Elang.Iprint e ->
    cfg_expr_of_eexpr f typ_fun e >>= fun e ->
    Hashtbl.replace cfg next (Cprint (e,succ));
    OK (next, next + 1)
  
  |Elang.Icall(fname,l_expr) -> cfg_expr_of_eexpr f typ_fun (Elang.Ecall(fname,l_expr)) >>= 
      fun e -> (match e with
                |Ecall(f,l) -> Hashtbl.replace cfg next (Ccall(f,l,succ)) ; OK(next,next+1)
                |_ -> Error "Ecall was expected ")
  
  | Elang.Istore(expr1,expr2) -> 
    cfg_expr_of_eexpr f typ_fun expr1 >>= fun e1 ->
    cfg_expr_of_eexpr f typ_fun expr2 >>= fun e2 ->
    type_expr f.funvartyp typ_fun (Eload(expr1)) >>= fun t ->
    Hashtbl.replace cfg next (Cstore(e1,e2,size_type t,succ)) ; OK(next,next+1)
      

(* Some nodes may be unreachable after the CFG is entirely generated. The
   [reachable_nodes n cfg] constructs the set of node identifiers that are
   reachable from the entry node [n]. *)
let rec reachable_nodes n (cfg: (int,cfg_node) Hashtbl.t) =
  let rec reachable_aux n reach =
    if Set.mem n reach then reach
    else let reach = Set.add n reach in
      match Hashtbl.find_option cfg n with
      | None -> reach
      | Some (Cnop succ)
      | Some (Cprint (_, succ))
      | Some (Cassign (_, _, succ)) -> reachable_aux succ reach
      | Some (Creturn _) -> reach
      | Some (Ccmp (_, s1, s2)) ->
        reachable_aux s1 (reachable_aux s2 reach)
      | Some(Ccall(_,_,succ)) -> reachable_aux succ reach
      |Some(Cstore(_,_,_,succ)) -> reachable_aux succ reach
  in reachable_aux n Set.empty

(* [cfg_fun_of_efun f] builds the CFG for E function [f]. *)
let cfg_fun_of_efun (f: Elang.efun) (typ_fun : (string, typ list * typ) Hashtbl.t) : cfg_fun res =
  let cfg = Hashtbl.create 17 in
  Hashtbl.replace cfg 0 (Creturn (Eint 0));
  cfg_node_of_einstr f typ_fun 1 cfg 0 f.funbody >>= fun (node, _) ->
  (* remove unreachable nodes *)
  let r = reachable_nodes node cfg in
  Hashtbl.filteri_inplace (fun k _ -> Set.mem k r) cfg;
  
  let l = List.map (fun (name_var,type_var) ->
    match Hashtbl.find_option f.funvarinmem name_var with
    |None -> Var(name_var)
    |Some(ofs) -> Addr(ofs,(size_type type_var))
    ) f.funargs in
  
  OK { cfgfunargs = l;
       cfgfunbody = cfg;
       cfgentry = node;
       cfgfunstksz = f.funstksz
     }

let cfg_gdef_of_edef gd (typ_fun : (string, typ list * typ) Hashtbl.t) =
  match gd with
    Gfun f -> cfg_fun_of_efun f typ_fun>>= fun f -> 
      OK (Gfun f)

let cfg_prog_of_eprog (ep: eprog) : cfg_fun prog res =
  let typ_fun = Hashtbl.create 1 in 
  List.iter(fun (fname,f) -> match f with |Gfun(f) -> Hashtbl.replace typ_fun fname (snd(List.split(f.funargs)),f.funrettype)) ep;
  assoc_map_res (fun fname gd -> cfg_gdef_of_edef gd typ_fun ) ep

let pass_cfg_gen ep =
  match cfg_prog_of_eprog ep with
  | Error msg ->
    record_compile_result ~error:(Some msg) "CFG"; Error msg
  | OK cfg ->
    record_compile_result "CFG";
    dump !cfg_dump dump_cfg_prog cfg (call_dot "cfg" "CFG");
    OK cfg
