open Ast
open Elang
open Prog
open Report
open Options
open Batteries
open Elang_print
open Utils

let tag_is_unop = 
  function
  |Tneg -> true
  |_ -> false

let unop_of_tag =
  function
    Tneg -> Eneg
    | _ -> assert false
let tag_is_binop =
  function
    Tadd -> true
  | Tsub -> true
  | Tmul -> true
  | Tdiv -> true
  | Tmod -> true
  | Txor -> true
  | Tcle -> true
  | Tclt -> true
  | Tcge -> true
  | Tcgt -> true
  | Tceq -> true
  | Tne  -> true
  | _    -> false

let binop_of_tag =
  function
    Tadd -> Eadd
  | Tsub -> Esub
  | Tmul -> Emul
  | Tdiv -> Ediv
  | Tmod -> Emod
  | Txor -> Exor
  | Tcle -> Ecle
  | Tclt -> Eclt
  | Tcge -> Ecge
  | Tcgt -> Ecgt
  | Tceq -> Eceq
  | Tne -> Ecne
  | _ -> assert false



(*[compatible_types t1 t2] is true if the types are compatible, false otherwise *)
let compatible_types (t1 : typ) (t2 : typ) :bool * typ =
  match t1,t2 with
  |t1',t2' when t1'=t2' -> true,t1'
  |Tchar,Tint -> true,Tchar
  |Tint,Tchar -> true,Tchar
  |Tptr(_),Tint -> true,t1
  |Tint,Tptr(_) -> true,t2
  |Tptr(_),Tchar -> true,t1
  |Tchar,Tptr(_) -> true,t2  
  |_,_ -> false,Tvoid



(*[type_and_name_of_node tree typ_var] donne le nom et le type de la varibale dans une definition ou assignation*)
let rec type_and_name_of_node (tree0 : tree) (typ_var : (string,typ) Hashtbl.t): (string * typ option) res =
  match tree0 with
  (*1 case : definitions*)
  |Node(t,[tree']) -> begin match t,tree' with
                      | Tint,StringLeaf(i) -> OK(i,Some(Tint))
                      | Tchar,StringLeaf(i) -> OK(i,Some(Tchar))
                      | Tvoid,StringLeaf(i) -> OK(i,Some(Tvoid))
                      | Tptr,_ -> type_and_name_of_node tree' typ_var >>= fun (i',t') ->
                        begin match t' with
                        |Some(t') -> OK(i',Some(Tptr(t')))
                        |None -> OK(i',None) end 
                      | _,_ -> Error (Format.sprintf "Error type : unknown tag : %s" (string_of_tag t)) end
  
  (*2 case : assignment : the variable has already been defined, so the type is not given again*)
  |StringLeaf(name_var) -> option_to_res_bind (Hashtbl.find_option typ_var name_var ) (Format.sprintf "Error type : type of variable %s was supposed to be known" name_var) (fun t -> OK(name_var,None))

  (*3 case : problem of type*)
  |_ -> Error (Format.sprintf "Unrecognised type in Assignvar : %s" (string_of_ast tree0))

(*[type_expr typ_var typ_fun e] détermine le type de l'expression e; renvoie un message d'erreur si l'expression n'est pas correctement typée *)
  let rec type_expr (typ_var : (string,typ) Hashtbl.t) (typ_fun : (string, typ list * typ) Hashtbl.t) (e: expr) : typ res =
    match e with

    | Echar(c) -> OK(Tchar)
    
    | Eint(i) -> OK(Tint)
    
    | Evar(name) -> option_to_res_bind (Hashtbl.find_option typ_var name) (Format.sprintf "Unrecognized variable '%s'" name) (fun t -> OK(t))
    
    |Ebinop(b,expr1,expr2) -> 
      type_expr typ_var typ_fun expr1 >>= fun t1 -> 
      type_expr typ_var typ_fun expr2 >>= fun t2 ->
        let bool,type_ret = compatible_types t1 t2 in
        if bool
          then OK(type_ret) 
          else Error (Format.sprintf "Error type : "^(dump_eexpr expr1)^" is of type "^(string_of_typ t1)^" while "^(dump_eexpr expr2)^" is of type "^(string_of_typ t2)) 
      
    |Eunop(u,expr) -> type_expr typ_var typ_fun expr
    
    | Ecall(fname,expr_list) -> option_to_res_bind (Hashtbl.find_option typ_fun fname) (Format.sprintf "%s is undefined" fname) (fun (typ_list,typ_ret) ->
      let rec aux expr_list typ_list typ_ret =
        match expr_list , typ_list with
        |[],[] -> OK(typ_ret)
        |eXpr::q,t'::q' -> type_expr typ_var typ_fun eXpr >>= fun t -> 
          if t = t' (*keep strong equality*)
            then aux q q' typ_ret 
            else Error(Format.sprintf "Error type : "^(dump_eexpr eXpr)^" is of type "^(string_of_typ t)^" while "^(string_of_typ t')^" was expected")
        |_,_ -> Error(Format.sprintf "Error type : %i arguments were expected, got %i" (List.length typ_list) (List.length expr_list))
      in aux expr_list typ_list typ_ret)
    
    | Eaddrof(expr0) -> type_expr typ_var typ_fun expr0 >>= fun t-> OK(Tptr(t))

    | Eload(expr0) -> type_expr typ_var typ_fun expr0 >>= fun t -> 
      match t with
      |Tptr(t') -> OK(t')
      |_ -> Error (Format.sprintf "Error type : type 'ptr' was expected but got %s" (string_of_typ t))





(* [make_eexpr_of_ast a] builds an expression corresponding to a tree [a]. If
   the tree is not well-formed, fails with an [Error] message. *)

let rec make_eexpr_of_ast (a: tree) (typ_var : (string,typ) Hashtbl.t) (typ_fun : (string, typ list * typ) Hashtbl.t) : expr res =
  let res =
    match a with

    (*value of a variable*)
    | StringLeaf(s) ->  type_expr typ_var typ_fun (Evar(s)) >>= fun _ -> OK(Evar(s))
    (*value of a character*)
    | Node(Tchar,[CharLeaf(c)]) -> type_expr typ_var typ_fun (Echar(c)) >>= fun _ -> OK(Echar(c))
    (*value of an integer*)
    | Node(Tint,[IntLeaf(i)]) -> type_expr typ_var typ_fun (Eint(i)) >>= fun _ -> OK(Eint(i))
    
    | Node(t, [e1; e2]) when tag_is_binop t -> 
      make_eexpr_of_ast e1 typ_var typ_fun >>= fun expr1 -> 
      make_eexpr_of_ast e2 typ_var typ_fun >>= fun expr2 -> 
      type_expr typ_var typ_fun (Ebinop(binop_of_tag t,expr1,expr2)) >>= fun _ ->
          OK(Ebinop(binop_of_tag t,expr1,expr2))
    
    | Node(t,[e]) when tag_is_unop t-> make_eexpr_of_ast e typ_var typ_fun >>= fun expr -> 
      type_expr typ_var typ_fun (Eunop(unop_of_tag t,expr)) >>= fun _ ->
      OK(Eunop(unop_of_tag t,expr))
    
    | Node(Tcall,[Node(Tfunname,[StringLeaf(f)]);Node(Targs,fargs)]) -> list_map_res (fun x -> make_eexpr_of_ast x typ_var typ_fun) fargs  >>= 
    fun l -> type_expr typ_var typ_fun (Ecall(f,l)) >>= fun _ ->
      OK(Ecall(f,l))

    |Node(Tptr,[tree]) -> make_eexpr_of_ast tree typ_var typ_fun >>= fun expr0 ->
      type_expr typ_var typ_fun (Eload(expr0)) >>= fun _ ->
        OK(Eload(expr0))
      
    |Node(Tamp,[tree]) -> make_eexpr_of_ast tree typ_var typ_fun >>= fun expr0 ->
      type_expr typ_var typ_fun (Eaddrof(expr0)) >>= fun _ ->
        OK(Eaddrof(expr0))

    | _ -> Error (Printf.sprintf "Unacceptable ast in make_eexpr_of_ast %s"
                    (string_of_ast a))
  in
  match res with
    OK o -> res
  | Error msg -> Error (Format.sprintf "In make_eexpr_of_ast %s:\n%s"
                          (string_of_ast a) msg)


let rec addr_taken_expr (e:expr) : string Set.t = 
  match e with
  | Ebinop (_,expr1,expr2) -> Set.union (addr_taken_expr expr1) (addr_taken_expr expr2)
  | Eunop (_,expr0) -> addr_taken_expr expr0
  | Eint(_) -> Set.empty
  | Echar(_) -> Set.empty
  | Evar(_) -> Set.empty
  | Ecall (_ ,expr_list) -> List.fold_left (fun acc e -> Set.union acc (addr_taken_expr e)) (Set.empty) expr_list
  | Eaddrof(expr0) -> (match expr0 with
                      |Evar(name_var) ->Set.singleton name_var
                      |_ -> Set.empty) (*<- pourra être modifié par la suite*)
  | Eload(expr0) -> addr_taken_expr expr0

let rec addr_taken_instr (i : instr) : string Set.t =
  match i with
  | Iassign(name_var,expr0) -> addr_taken_expr expr0
  | Iif(expr0,instr1,instr2) -> Set.union (addr_taken_expr expr0) (Set.union (addr_taken_instr instr1) (addr_taken_instr instr2))
  | Iwhile(expr0,instr0) -> Set.union (addr_taken_expr expr0) (addr_taken_instr instr0)
  | Iblock(instr_list) -> List.fold_left (fun acc instr0 -> Set.union acc (addr_taken_instr instr0)) (Set.empty) instr_list
  | Ireturn(expr0) -> addr_taken_expr expr0
  | Iprint(expr0) -> addr_taken_expr expr0 
  | Istore(expr1,expr2) -> addr_taken_expr expr2
  | Icall(fname,expr_list) -> List.fold_left (fun acc e -> Set.union acc (addr_taken_expr e)) (Set.empty) expr_list


let rec make_einstr_of_ast (a: tree) (typ_var : (string,typ) Hashtbl.t) (typ_fun : (string, typ list * typ) Hashtbl.t) : instr res =
  let res =
    match a with
    (*VARIABLE : definition with value*)
    | Node(Tassign,[Node(Tassignvar,[Node(t,[StringLeaf(i)]);e])]) when t != Tptr -> 
        make_eexpr_of_ast e typ_var typ_fun >>= fun expr -> 
        type_expr typ_var typ_fun expr >>= fun t' ->
        typ_of_tag t >>= fun t ->
          let bool,type_ret = compatible_types t t' in
          if bool
            then begin Hashtbl.replace typ_var i type_ret;  OK(Iassign(i,expr)) end
            else Error(Format.sprintf "Uncompatible types : type '%s' of '%s' can not match with type '%s' of expression %s"  (string_of_typ t) i (string_of_typ t') (dump_eexpr expr)) 
    
    (*VARIABLE : definition without value*)
    |Node(Tassign,[Node(Tassignvar,[Node(t,[StringLeaf(i)])])]) when t != Tptr-> 
      typ_of_tag t >>= fun t ->
        begin Hashtbl.replace typ_var i t;  
         match t with 
         | Tint -> OK(Iassign(i,Eint(0))) 
         | Tchar -> OK(Iassign(i,Echar('0'))) 
         | Tvoid -> Error "Variable can not fit with type void"
         | Tptr _ -> Error "Impossible to get Tptr here !"
        end

    (*POINTER : definition/Assignment with value*)
    | Node(Tassign,[Node(Tassignvar,[Node(Tptr,[e1]);e2])]) -> 
      type_and_name_of_node e1 typ_var >>= fun (name_var,type_var_opt) ->
        begin match type_var_opt with
        (*Definition : because type is indicated *)
        |Some(type_var) -> 
          Hashtbl.replace typ_var name_var (Tptr(type_var)); 
          make_eexpr_of_ast e2 typ_var typ_fun >>= fun expr2 ->
          OK(Iassign(name_var,expr2))
        (*Assignment : no type is precised, only Asterixs (Tptr)*)
          |None -> 
            (*MANQUE CONTROLE DE TYPE*)
            make_eexpr_of_ast e1 typ_var typ_fun >>= fun expr1 ->
            make_eexpr_of_ast e2 typ_var typ_fun >>= fun expr2 -> 
            OK(Istore(expr1,expr2))
      end

    (*POINTER : definition without value*)
    |Node(Tassign,[Node(Tassignvar,[Node(Tptr,[e1])])]) ->  
      type_and_name_of_node e1 typ_var >>= fun (name_var,type_var) ->
        begin match type_var with
        |Some(t) -> 
          Hashtbl.replace typ_var name_var (Tptr(t)); 
          OK(Iassign(name_var,Eint(0)))
        |None -> Error "Error type : type is missing for a definition"
      end


    (*Assignation de varibale déjà définie*)    
    | Node(Tassign,[Node(Tassignvar,[StringLeaf(i);e])]) ->
      type_expr typ_var typ_fun (Evar(i)) >>= fun t -> 
      make_eexpr_of_ast e typ_var typ_fun >>= fun expr ->
      type_expr typ_var typ_fun expr >>= fun t' ->
        let bool,type_ret = compatible_types t t' in
        if bool 
          then begin Hashtbl.replace typ_var i type_ret;  OK(Iassign(i,expr)) end
          else Error(Format.sprintf "Uncompatible types : type '%s' of '%s' can not match with type '%s' of expression %s" (string_of_typ t) i (string_of_typ t') (dump_eexpr expr)) 
        

    | Node(Tif,[e;i1;i2]) -> 
      make_eexpr_of_ast e typ_var typ_fun >>= fun expr -> 
        make_einstr_of_ast i1 typ_var typ_fun>>= fun instr1 -> 
        make_einstr_of_ast i2 typ_var typ_fun>>= fun instr2 -> 
        OK(Iif(expr,instr1,instr2))
    
    | Node(Twhile,[e;i1]) -> 
      make_eexpr_of_ast e typ_var typ_fun>>= fun expr -> 
        make_einstr_of_ast i1 typ_var typ_fun >>= fun instr1 -> 
          OK(Iwhile(expr,instr1))
      
    | Node(Treturn,[e]) -> 
      make_eexpr_of_ast e typ_var typ_fun >>= fun expr -> OK(Ireturn(expr))

    | Node(Tprint,[e]) -> 
      make_eexpr_of_ast e typ_var typ_fun >>= fun expr -> OK(Iprint(expr))
    
    | Node(Tblock,l) -> 
      (list_map_res (fun x-> make_einstr_of_ast x typ_var typ_fun) l )>>=(fun linstr -> OK(Iblock(linstr)))

    | Node(Tcall,[Node(Tfunname,[StringLeaf(f)]);Node(Targs,fargs)]) -> 
      list_map_res (fun x -> make_eexpr_of_ast x typ_var typ_fun) fargs >>= fun l -> 
        type_expr typ_var typ_fun (Ecall(f,l)) >>= fun _ ->
          OK(Icall(f,l))

    | _ -> Error (Printf.sprintf "Unacceptable ast in make_einstr_of_ast %s"
                    (string_of_ast a))
  in
  match res with
    OK o -> res
  | Error msg -> Error (Format.sprintf "In make_einstr_of_ast %s:\n%s"
                          (string_of_ast a) msg)
                      



let make_ident (a: tree) (typ_var : (string,typ) Hashtbl.t) : (string * typ) res =
  match a with
  | Node (Targ, [a']) -> type_and_name_of_node a' typ_var >>=
  fun(name_var,type_var_opt) -> 
    option_to_res_bind type_var_opt "Error type : type of variable is missing for a definiton" (fun t -> OK(name_var,t))

  | a -> Error (Printf.sprintf "make_ident: unexpected AST: %s"
                  (string_of_ast a))

let make_fundef_of_ast (a: tree) (typ_fun : (string, typ list * typ) Hashtbl.t) : (string * efun)  res =
  let typ_var = Hashtbl.create 10  in 
  match a with
  (* TODO *)
  (*pre-definition for mutually recursive calls*)
  | Node(Tfundef,[Node(Tfunname,[Node(t,[StringLeaf(fname)])]);Node(Tfunargs,fargs);Node(Tfunbody,[])]) ->
    typ_of_tag t >>= fun t ->
      list_map_res (fun x -> make_ident x typ_var) fargs >>= fun lvar -> 
        Hashtbl.replace typ_fun fname (snd(List.split lvar),t);
        List.iter (fun (n_var,t_var) -> Hashtbl.replace typ_var n_var t_var) lvar;
        OK(fname,{funargs = lvar; funbody = Ireturn(Echar('0')) ; funvartyp = typ_var; funrettype = t ; funvarinmem = Hashtbl.create 0 ; funstksz = 0})
  
  (*definition of function*)
  | Node(Tfundef,[Node(Tfunname,[Node(t,[StringLeaf(fname)])]);Node(Tfunargs,fargs);Node(Tfunbody,[fbody])]) -> 
      typ_of_tag t >>= fun t ->
      list_map_res (fun x -> make_ident x typ_var) fargs >>= fun lvar -> 
        begin match Hashtbl.find_option typ_fun fname with
        
          (*Case 1 : the fonction was previously defined*)
          |Some(type_var',ret_type') -> if (t = ret_type') && (type_var' = (snd(List.split lvar))) 
            then begin Hashtbl.replace typ_fun fname (snd(List.split lvar),t);
                List.iter (fun (name_var,type_var) -> Hashtbl.replace typ_var name_var type_var) lvar;
                make_einstr_of_ast fbody typ_var typ_fun >>= fun instr ->
                let funvarinmem = Hashtbl.create 1 in
                let ofs = Set.fold (fun name_var ofs -> Hashtbl.replace funvarinmem name_var ofs ; ofs+(size_type(Hashtbl.find typ_var name_var))) 
                  (addr_taken_instr instr ) 0 in
                  OK(fname,{funargs = lvar; funbody = instr ; funvartyp = typ_var; funrettype = t ; funvarinmem = funvarinmem ; funstksz = ofs}) end
            
            
            else Error(Format.sprintf "New definition of %s does not match with previous definition : 
            previous return type was %s, current return type is %s ; 
            previous arguments types were %s, current arguments types are %s"  
            fname (string_of_typ ret_type') (string_of_typ t) 
            (List.fold_right (fun t acc -> (string_of_typ t)^(" ")^acc) type_var' "")
            (List.fold_right (fun t acc -> (string_of_typ t)^(" ")^acc) (snd(List.split lvar)) "")
            )
          (*Case 2 : the fonction has never been defined*)
          |None ->
            Hashtbl.replace typ_fun fname (snd(List.split lvar),t);
            List.iter (fun (name_var,type_var) -> Hashtbl.replace typ_var name_var type_var) lvar;
            make_einstr_of_ast fbody typ_var typ_fun >>= fun instr ->
              let funvarinmem = Hashtbl.create 1 in
              let ofs = Set.fold (fun name_var ofs -> Hashtbl.replace funvarinmem name_var ofs ; ofs+(size_type(Hashtbl.find typ_var name_var))) 
              (addr_taken_instr instr ) 0 in
              OK(fname,{funargs = lvar; funbody = instr ; funvartyp = typ_var; funrettype = t ; funvarinmem = funvarinmem ; funstksz = ofs}) end
      

  | _ -> Error (Printf.sprintf "make_fundef_of_ast: Expected a Tfundef, got %s."
             (string_of_ast a))

let make_eprog_of_ast (a: tree) : eprog res =
  let fun_typ = Hashtbl.create (10) in
  Hashtbl.replace fun_typ "print" ([Tint], Tvoid);
  Hashtbl.replace fun_typ "print_int" ([Tint], Tvoid);
  Hashtbl.replace fun_typ "print_char" ([Tchar], Tvoid);
  match a with
  | Node (Tlistglobdef, l) ->
    list_map_res (fun a -> make_fundef_of_ast a fun_typ >>= fun (fname, efun) -> OK (fname, Gfun efun)) l >>= 
    (* only save the last definition of a function, e.g previous definitions come from recursive definitions *)  
    fun r -> let rec aux l =
        match l with
        |[] -> []
        |(s,f)::q -> match List.assoc_opt s q with
                      |None -> (s,f)::aux q
                      |Some(_) -> aux q 
      in OK(aux r) 
  | _ ->
    Error (Printf.sprintf "make_fundef_of_ast: Expected a Tlistglobdef, got %s."
             (string_of_ast a))

let pass_elang ast =
  match make_eprog_of_ast ast with
  | Error msg ->
    record_compile_result ~error:(Some msg) "Elang";
    Error msg
  | OK  ep ->
    dump !e_dump dump_e ep (fun file () ->
        add_to_report "e" "E" (Code (file_contents file))); OK ep

