open Elang
open Elang_gen
open Elang_print
open Batteries
open BatList
open Prog
open Utils
open Builtins
open Utils


let binop_bool_to_int f x y = if f x y then 1 else 0

(* [eval_binop b x y] évalue l'opération binaire [b] sur les arguments [x]
   et [y]. *)
let eval_binop (b: binop) : int -> int -> int =
  match b with
  | Eadd -> fun x y -> x+y
  | Emul -> fun x y -> x*y
  | Emod -> fun x y -> x mod y
  | Exor -> fun x y -> 0 (*pas code*)
  | Ediv -> fun x y -> x/y
  | Esub -> fun x y -> x-y
  | Eclt -> binop_bool_to_int (fun x y -> x<y)
  | Ecle -> binop_bool_to_int (fun x y -> x<=y)
  | Ecgt -> binop_bool_to_int (fun x y -> x>y)
  | Ecge -> binop_bool_to_int (fun x y -> x>=y)
  | Eceq -> binop_bool_to_int (fun x y -> x==y)
  | Ecne -> binop_bool_to_int (fun x y -> x!=y)


(* [eval_unop u x] évalue l'opération unaire [u] sur l'argument [x]. *)
let eval_unop (u: unop) : int -> int =
  match u with
   | Eneg -> fun x -> -x

(* [eval_eexpr st e] évalue l'expression [e] dans l'état [st]. Renvoie une
   erreur si besoin. *)
let rec eval_eexpr oc (ep: eprog) (st: int state) (sp : int) (typ_fun : (string, typ list * typ) Hashtbl.t) (f : efun) (e : expr)  : int res =
   (*TODO*)
   match e with
   
   | Echar(c) -> OK(Char.code(c))
  
   | Ebinop (binop,expr1,expr2) ->  
      eval_eexpr oc ep st sp typ_fun f expr1 >>= fun int1 ->  
      eval_eexpr oc ep st sp typ_fun f expr2 >>= fun int2 -> 
      type_expr f.funvartyp typ_fun expr1 >>= fun t1 ->
      type_expr f.funvartyp typ_fun expr2 >>= fun t2 ->
      begin match t1,t2 with 
         |Tptr _,Tint when binop = Eadd-> OK(eval_binop Esub int1 ((size_type t1)*int2))

         |Tint,Tptr _ when binop = Eadd -> OK(eval_binop Esub int2 ((size_type t2)*int1) )

         |Tptr _,Tint when binop = Esub-> OK(eval_binop Eadd int1 ((size_type t1)*int2))

         |_,_  -> OK(eval_binop binop int1 int2)

      end
  
      
 
   | Eunop (unop,expr) -> eval_eexpr oc ep st sp typ_fun f expr >>= fun int0 ->  
      OK(eval_unop unop int0)
   
   | Eint(i) -> OK(i)
   
   (*A variable can be either in [env] or in [mem]*)
   | Evar(name_var) -> 
      begin match (Hashtbl.find_option st.env name_var) with
         | Some(value_var) -> OK(value_var)
         | None -> begin match Hashtbl.find_option f.funvarinmem name_var with
                           | Some(ofs_var) -> type_expr f.funvartyp typ_fun e >>= fun t ->
                              Mem.read_bytes_as_int (st.mem) (sp-ofs_var) (size_type t) 
                           | None -> Error (Format.sprintf "Unknown variable %s in environment or memory" name_var)
                  end 
         end

   | Ecall(fname,expr_args) ->  list_map_res (fun eXpr -> eval_eexpr oc ep st sp typ_fun f eXpr) expr_args >>= 
      fun vargs ->  begin match find_function ep fname with
      |OK(f') -> eval_efun oc ep st (sp-f.funstksz) typ_fun f' fname vargs >>= 
            fun (ret,st') ->  option_to_res_bind ret "no return" (fun ret -> OK(ret))
      |Error(s) ->  do_builtin oc st.mem fname vargs >>= 
            fun r -> begin match r with
            |None -> Error (Printf.sprintf "%s ne retourne rien " fname)
            |Some(i) -> OK(i) end end
   
   | Eaddrof(e) -> begin match e with
                        | Evar(name_var) -> option_to_res_bind (Hashtbl.find_option f.funvarinmem name_var) (Format.sprintf "%s is not in memory" name_var) (fun ofs -> OK(sp-ofs))
                        | _ -> Error (Format.sprintf "Can only calculate the address of a varibale, got %s" (dump_eexpr e)) 
                        end
   
   | Eload(e') -> 
      eval_eexpr oc ep st sp typ_fun f e' >>= fun addr -> 
         type_expr f.funvartyp typ_fun e >>= fun t ->
         Mem.read_bytes_as_int (st.mem) (addr) (size_type t)
   


(* [eval_einstr oc st ins] évalue l'instrution [ins] en partant de l'état [st].

   Le paramètre [oc] est un "output channel", dans lequel la fonction "print"
   écrit sa sortie, au moyen de l'instruction [Format.fprintf].

   Cette fonction renvoie [(ret, st')] :

   - [ret] est de type [int option]. [Some v] doit être renvoyé lorsqu'une
   instruction [return] est évaluée. [None] signifie qu'aucun [return] n'a eu
   lieu et que l'exécution doit continuer.

   - [st'] est l'état mis à jour. *)
and eval_einstr oc (ep : eprog) (st: int state) (sp : int) (typ_fun : (string, typ list * typ) Hashtbl.t) (f : efun) (ins: instr) :
  (int option * int state) res =
  match ins with
  
 
  | Iassign(name_var,expr) -> 
      eval_eexpr oc ep st sp typ_fun f expr >>= fun val_expr -> 
      begin match Hashtbl.find_option f.funvarinmem name_var with
         |Some(ofs_var) -> type_expr f.funvartyp typ_fun (Evar(name_var)) >>= fun t -> (* variable has to be written in memory*)
         (*failwith (Format.sprintf "Iassign : %s <- %i \n %s" name_var val_expr (string_of_funvarinmem f.funvarinmem))*)
         Mem.write_bytes (st.mem) (sp-ofs_var) (split_bytes (size_type t) val_expr) >>= fun () ->
               OK(None,st)
         | None -> Hashtbl.replace st.env name_var val_expr; OK(None,st) (*otherwise we write variable in env*)
      end

   
  
   | Iif(expr,instr1,instr2) -> eval_eexpr oc ep st sp typ_fun f expr >>= (fun int -> if int!=0 then eval_einstr oc ep st sp typ_fun f instr1 else eval_einstr oc ep st sp typ_fun f instr2)

   | Iwhile(expr,instr) -> eval_eexpr oc ep st sp typ_fun f expr >>= (fun val_expr -> 
      if val_expr !=0 
         then ( eval_einstr oc ep st sp typ_fun f instr >>= fun (in_opt,sT) -> 
            (match in_opt with
            | Some(_) ->OK(in_opt,sT)
            | None -> eval_einstr oc ep sT sp typ_fun f ins) ) 
         else OK(None,st))
  
  | Iblock(linstr) -> (match linstr with
                        |[] -> OK(None,st)
                        |p::q -> eval_einstr oc ep st sp typ_fun f p >>= (fun (in_opt,sT) ->
                           (match in_opt with
                                    | Some(_) ->OK(in_opt,sT)
                                    | None -> eval_einstr oc ep sT sp typ_fun f (Iblock(q)))))

  | Ireturn(expr) -> eval_eexpr oc ep st sp typ_fun f expr >>= fun int -> OK(Some(int),st)
  
  | Iprint(expr) -> eval_eexpr oc ep st sp typ_fun f expr >>= (fun i -> (Format.fprintf oc "%d\n" i); OK(None,st))
  
  | Icall(fname,expr_args) ->  list_map_res (fun eXpr -> eval_eexpr oc ep st sp typ_fun f eXpr) expr_args >>= 
            fun vargs -> begin match find_function ep fname with
            |OK(f') -> eval_efun oc ep st (sp-f.funstksz) typ_fun f' fname vargs >>= 
                  fun (ret,st') ->  OK(None,st)
            |Error(s) ->  do_builtin oc st.mem fname vargs >>= 
                  fun r -> OK(r,st) end
  
   |Istore(expr1,expr2) -> 
      eval_eexpr oc ep st sp typ_fun f expr1 >>= fun addr_expr1 ->
      eval_eexpr oc ep st sp typ_fun f expr2 >>= fun val_expr2 ->
      type_expr f.funvartyp typ_fun expr1 >>= fun t -> 
         (*failwith (Format.sprintf "testIstore" )*)
      Mem.write_bytes (st.mem) (addr_expr1) (split_bytes (size_type t) val_expr2)>>= fun () ->
         OK(None,st)  

      
  (* [eval_efun oc st f fname vargs] évalue la fonction [f] (dont le nom est
   [fname]) en partant de l'état [st], avec les arguments [vargs].

   Cette fonction renvoie un couple (ret, st') avec la même signification que
   pour [eval_einstr]. *)
and eval_efun oc ep (st: int state) (sp : int) (typ_fun : (string, typ list * typ) Hashtbl.t) ({funargs; funbody;funvartyp;funrettype;funvarinmem;funstksz} as f: efun)
    (fname: string) (vargs: int list)
  : (int option * int state) res =
  (* L'environnement d'une fonction (mapping des variables locales vers leurs
     valeurs) est local et un appel de fonction ne devrait pas modifier les
     variables de l'appelant. Donc, on sauvegarde l'environnement de l'appelant
     dans [env_save], on appelle la fonction dans un environnement propre (Avec
     seulement ses arguments), puis on restore l'environnement de l'appelant. *)
  let env_save = Hashtbl.copy st.env in
  let env = Hashtbl.create 17 in
  match List.iter2 (fun name_var val_var -> begin match Hashtbl.find_option f.funvarinmem name_var with
         |None -> Hashtbl.replace env name_var val_var
         |Some(ofs_var) -> (match (type_expr f.funvartyp typ_fun (Evar(name_var)) >>= fun t ->
            Mem.write_bytes (st.mem) (sp-ofs_var) (split_bytes (size_type t) val_var)) with
            |Error s -> failwith s
            |_->() )
         end )
  ( fst (List.split funargs)) vargs with
  | () ->
   
    eval_einstr oc ep { st with env } sp typ_fun f funbody >>= fun (v, st') ->
    OK (v, { st' with env = env_save })
  | exception Invalid_argument _ ->
    Error (Format.sprintf
             "E: Called function %s with %d arguments, expected %d.\n"
             fname (List.length vargs) (List.length funargs)
          )

(* [eval_eprog oc ep memsize params] évalue un programme complet [ep], avec les
   arguments [params].

   Le paramètre [memsize] donne la taille de la mémoire dont ce programme va
   disposer. Ce n'est pas utile tout de suite (nos programmes n'utilisent pas de
   mémoire), mais ça le sera lorsqu'on ajoutera de l'allocation dynamique dans
   nos programmes.

   Renvoie:

   - [OK (Some v)] lorsque l'évaluation de la fonction a lieu sans problèmes et renvoie une valeur [v].

   - [OK None] lorsque l'évaluation de la fonction termine sans renvoyer de valeur.

   - [Error msg] lorsqu'une erreur survient.
   *)
let eval_eprog oc (ep: eprog) (memsize: int) (params: int list)
  : int option res =
  let st = init_state memsize in
  let typ_fun = Hashtbl.create 1 in 
  List.iter(fun (fname,f) -> match f with |Gfun(f) -> Hashtbl.replace typ_fun fname (snd(List.split(f.funargs)),f.funrettype)) ep;
  let sp = memsize-4 in
  find_function ep "main" >>= fun f ->
  (* ne garde que le nombre nécessaire de paramètres pour la fonction "main". *)
  let n = List.length f.funargs in
  let params = take n params in
  eval_efun oc ep st sp typ_fun f "main" params >>= fun (v, st) ->
  OK v
